from typing import TypeVar

from aiocdp.core.interfaces.stream import IEventStreamReader, IEventStream
from aiocdp.core.interfaces.connection import IConnection
from aiocdp.core.interfaces.session import ISession
from aiocdp.core.interfaces.chrome import IChrome
from aiocdp.core.interfaces.target import ITarget, ITargetInfo
from aiocdp.utils import UNDEFINED

_T = TypeVar('_T')
_lookup = {}


def get_class(interface: type[_T], default: type = UNDEFINED) -> type[_T]:
    if interface in _lookup:
        return _lookup[interface]

    if default is UNDEFINED:
        raise RuntimeError(
            f'No implementation registered for interface: {interface}'
        )

    assert issubclass(
        default,
        interface
    )

    return default


def set_class(interface: type, implementation: type):
    assert issubclass(
        implementation,
        interface
    ), f'Implementation {implementation} must be a subclass of interface {interface}'

    _lookup[interface] = implementation


def set_chrome_class(implementation: type[IChrome]):
    return set_class(
        IChrome,
        implementation
    )


def set_target_class(implementation: type[ITarget]):
    return set_class(
        ITarget,
        implementation
    )


def set_target_info_class(implementation: type[ITargetInfo]):
    return set_class(
        ITargetInfo,
        implementation
    )


def set_session_class(implementation: type[ISession]):
    return set_class(
        ISession,
        implementation
    )


def set_connection_class(implementation: type[IConnection]):
    return set_class(
        IConnection,
        implementation
    )


def set_event_stream_class(implementation: type[IEventStream]):
    return set_class(
        IEventStream,
        implementation
    )


def set_event_stream_reader_class(implementation: type[IEventStreamReader]):
    return set_class(
        IEventStreamReader,
        implementation
    )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, TYPE_CHECKING, AsyncIterator, Iterator

if TYPE_CHECKING:
    from aiocdp.core.interfaces.connection import IConnection

_T = TypeVar('_T')


@dataclass
class IEventStream(ABC):
    """
    Represents an asynchronous stream of CDP events received from the connection.
    """

    @classmethod
    @abstractmethod
    def init(
        cls,
        connection: 'IConnection',
        events_to_listen: list[str]
    ) -> 'IEventStream':
        """
        Initializer method for the IEventStream class.
        """
        pass

    @abstractmethod
    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Unsubscribes this stream from receiving events from the connection.
        """
        pass

    @abstractmethod
    def get_connection(self) -> 'IConnection':
        """
        Returns the connection this event stream is associated with.
        """
        pass

    @abstractmethod
    def get_events_to_listen(self) -> list[str]:
        """
        Returns a list of events to listen to.
        """
        pass

    @abstractmethod
    def get_reader(self) -> 'IEventStreamReader':
        """
        Returns a read only interface associated with this event stream.
        """
        pass

    @abstractmethod
    def is_closed(self) -> bool:
        """
        Returns whether the stream is closed.
        """
        pass

    @abstractmethod
    def iterate(self) -> AsyncIterator[_T]:
        """
        Returns an async iterator for all recorded events and new events as they are received.
        """
        pass

    @abstractmethod
    def iterate_already_recorded(self) -> Iterator[_T]:
        """
        Returns an async iterator for all recorded events.
        """
        pass

    @abstractmethod
    def write(self, event: '_T'):
        """
        Writes an event to the stream.
        """
        pass


class IEventStreamReader(ABC):
    """
    Read only interface to an event stream.
    """

    @abstractmethod
    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Closes this stream.
        """
        pass

    @abstractmethod
    def get_connection(self) -> 'IConnection':
        """
        Returns the connection this event stream is associated with.
        """
        pass

    @abstractmethod
    def get_events_to_listen(self):
        """
        Returns a list of events to listen to.
        """
        pass

    @abstractmethod
    def is_closed(self) -> bool:
        """
        Returns whether the stream is closed.
        """
        pass

    @abstractmethod
    def iterate(self) -> AsyncIterator[_T]:
        """
        Returns an async iterator for all recorded events and new events as they are received.
        """
        pass

    @abstractmethod
    def iterate_already_recorded(self) -> Iterator[_T]:
        """
        Returns an async iterator for all recorded events.
        """
        pass



from .core.implementations import (
    Chrome,
    Target,
    Session,
    Connection,
    EventStream,
    EventStreamReader,
    TargetInfo
)
from .ioc import (
    set_chrome_class,
    set_target_class,
    set_target_info_class,
    set_session_class,
    set_connection_class,
    set_event_stream_class,
    set_event_stream_reader_class
)


def setup_default_factories():
    set_chrome_class(Chrome)
    set_target_class(Target)
    set_target_info_class(TargetInfo)
    set_session_class(Session)
    set_connection_class(Connection)
    set_event_stream_class(EventStream)
    set_event_stream_reader_class(EventStreamReader)



from abc import ABC, abstractmethod
from asyncio import Future
from typing import TYPE_CHECKING, Any, Coroutine

from aiocdp.core.interfaces.session import ISession
from aiocdp.core.interfaces.stream import IEventStream, IEventStreamReader

if TYPE_CHECKING:
    from aiocdp.core.interfaces.chrome import IChrome


class ITargetInfo(ABC):
    """
    Represents information about a CDP target.
    """

    @classmethod
    @abstractmethod
    def init(
        cls,
        id_: str,
        title: str,
        description: str,
        url: str,
        type_: str,
        web_socket_debugger_url: str
    ):
        """
        Initializer method for ITargetInfo subclasses
        """
        pass

    @abstractmethod
    def get_id(self) -> str:
        """
        Returns the id of the target.
        """
        pass


class ITarget(ABC):
    """
    Represents a CDP target.
    """

    @classmethod
    @abstractmethod
    def init(
        cls,
        chrome: 'IChrome',
        info: 'ITargetInfo'
    ):
        """
        Initializer method for ITarget subclasses
        """
        pass

    @abstractmethod
    def close_session(self, session: 'ISession') -> Coroutine[None, None, Any]:
        """
        Closes the given session.
        """
        pass

    @abstractmethod
    def close_stream(self, stream: 'IEventStream') -> None:
        """
        Closes the given stream.
        """
        pass

    @abstractmethod
    def connect(self) -> Coroutine[None, None, Any]:
        """
        Connects to the target.
        """
        pass

    @abstractmethod
    def disconnect(self) -> Coroutine[None, None, Any]:
        """
        Disconnects from the target.
        """
        pass

    @abstractmethod
    def get_info(self) -> 'ITargetInfo':
        """
        Returns the associated ITargetInfo implementation
        """
        pass

    @abstractmethod
    def get_ws_url(self) -> str:
        """
        Returns the websocket url of the target.
        """
        pass

    @abstractmethod
    def is_connected(self) -> bool:
        """
        Returns whether the target is connected.
        """
        pass

    @abstractmethod
    def open_stream(self, events) -> IEventStreamReader:
        """
        Opens a stream for the given events.
        """
        pass

    @abstractmethod
    def open_session(self) -> Coroutine[None, None, 'ISession']:
        """
        Opens a session for the target.
        """
        pass

    @abstractmethod
    def send(self, method, params) -> Coroutine[None, None, Any]:
        """
        Sends a message to the target.
        """
        pass

    @abstractmethod
    def send_and_await_response(self, method, params) -> Coroutine[None, None, Any]:
        """
        Sends a message to the target and awaits a response.
        """
        pass



import subprocess
from dataclasses import dataclass
from typing import Callable, Any, Self

import requests

from aiocdp.core.implementations.target import Target, TargetInfo
from aiocdp.core.interfaces.chrome import IChrome
from aiocdp.core.interfaces.target import ITarget, ITargetInfo
from aiocdp.exceptions import NoTargetFoundMatchingCondition
from aiocdp.ioc import get_class
from aiocdp.utils import UNDEFINED


@dataclass
class Chrome(IChrome[subprocess.Popen]):
    """
    Represents a chrome instance.
    """

    """
    The host of the chrome instance.
    """
    host: str = '127.0.0.1'

    """
    The port of the chrome instance.
    """
    port: int = 9222

    @classmethod
    def init(
            cls,
            host: str = '127.0.0.1',
            port: int = 9222,
    ) -> 'Chrome':
        """
        Initializer method for the Chrome classes
        """
        return cls(
            host=host,
            port=port
        )

    def _init_target(self, target: dict):
        """
        Initializes an instance of the target using the registered implementations.
        """
        target_info_cls = get_class(
            ITargetInfo,
            TargetInfo
        )

        target_cls = get_class(
            ITarget,
            Target
        )

        target_info = target_info_cls.init(
            id_=target['id'],
            title=target['title'],
            description=target['description'],
            url=target['url'],
            type_=target['type'],
            web_socket_debugger_url=target['webSocketDebuggerUrl']
        )

        target = target_cls.init(
            chrome=self,
            info=target_info
        )

        return target

    def start(
            self,
            start_command: str = 'start chrome',
            extra_cli_args: list[str] = None,
            popen_kwargs: dict = None
    ) -> subprocess.Popen:
        """
        Starts chrome through the command line. Returns subprocess.Popen object.
        """
        commands = [
            start_command,
            f'--remote-debugging-port={self.port}',
            *(extra_cli_args or [])
        ]

        command = ' '.join(commands)

        return subprocess.Popen(
            command,
            **(popen_kwargs or {})
        )

    def close_tab(self, target_id: str):
        """
        Closes the tab with the given target id.
        """
        url = f'http://{self.host}:{self.port}/json/close/{target_id}'

        response = requests.get(url)
        response.raise_for_status()

    def get_first_target(
            self,
            condition: Callable[[ITarget], bool] = None,
            default: Any = UNDEFINED
    ) -> ITarget:
        """
        Fetches and returns the first target.
        Raises an exception if no target is found and no default is supplied.
        """
        for target in self.iterate_targets():
            if not condition:
                return target

            if condition(target):
                return target

        if default is not UNDEFINED:
            return default

        raise NoTargetFoundMatchingCondition(
            condition
        )

    def get_host(self) -> str:
        """
        Returns the host of the chrome instance
        """
        return self.host

    def get_port(self) -> int:
        """
        Returns the port of the chrome instance
        """
        return self.port

    def get_targets(self) -> list[ITarget]:
        """
        Fetches a list of all the targets.
        """
        return list(
            self.iterate_targets()
        )

    def iterate_targets(self):
        """
        Returns a generator that fetches and iterates over all the targets.
        """
        url = f'http://{self.host}:{self.port}/json/list'

        response = requests.get(url)
        response.raise_for_status()

        for target in response.json():
            yield self._init_target(target)

    def open_tab(self, url: str = None) -> ITarget:
        """
        Opens a new tab.
        """
        url_ = f'http://{self.host}:{self.port}/json/new'

        if url:
            url_ += f'?{url}'

        response = requests.put(url_)
        response.raise_for_status()

        target = response.json()

        return self._init_target(target)



import asyncio
from dataclasses import dataclass, field
from typing import TypeVar, Generator, AsyncIterator, Iterator

from aiocdp.core.interfaces.connection import IConnection
from aiocdp.core.interfaces.stream import IEventStream, IEventStreamReader

_T = TypeVar('_T')


def _create_future():
    """
    Utility function to create a new future.
    """
    return asyncio.get_event_loop().create_future()


@dataclass(eq=False)
class EventStream(IEventStream):
    """
    Represents an asynchronous stream of CDP events received from the connection.
    """

    """
    A reference to the parent connection.
    """
    connection: 'IConnection'

    """
    The list of event names to listen to.
    """
    events_to_listen: list[str]

    """
    The list of recorded events.
    """
    events: list[_T] = field(
        default_factory=list,
        init=False
    )

    """
    References the single instance of the reader for this class.
    """
    reader: 'EventStreamReader | None' = field(
        default=None,
        init=False
    )

    """
    An async future resolved everytime an event is recorded.
    """
    next: asyncio.Future = field(
        default_factory=_create_future,
        init=False
    )

    @classmethod
    def init(
        cls,
        connection: IConnection,
        events_to_listen: list[str]
    ) -> 'EventStream':
        """
        Initializer method for the EventStream class
        """
        return cls(
            connection,
            events_to_listen
        )

    def is_closed(self):
        """
        Public readonly access to the closed status of the stream.
        """
        return self.connection.is_stream_closed(self)

    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        self.close()

        if exc_type is not None:
            return False

    def close(self):
        """
        Unsubscribes this stream from receiving events from the connection.
        """
        self.connection.close_stream(self)

    def get_connection(self) -> 'IConnection':
        """
        Returns the connection this event stream is associated to.
        """
        return self.connection

    def get_events_to_listen(self):
        """
        Returns the list of events this stream is subscribed to
        """
        return self.events_to_listen

    def get_reader(self):
        """
        Returns a read only interface associated with this event stream.

        Implementation Notes:
        - Returns the same instance for the lifetime of this stream.
        """

        if self.reader is None:
            self.reader = EventStreamReader(self)

        return self.reader

    async def iterate(self):
        """
        Returns a new async iterator for all recorded events and new events as they are received.
        """
        for item in self.events:
            yield item

        while True:
            yield await self.next

    def iterate_already_recorded(self) -> Iterator[_T]:
        """
        Returns an async iterator for all recorded events.
        """
        for item in self.events:
            yield item

    def write(self, item: _T):
        """
        Writes an item to the stream. Responsible for resolving futures.
        """
        self.events.append(item)
        self.next.set_result(item)
        self.next = _create_future()


@dataclass
class EventStreamReader(IEventStreamReader):
    """
    Read only proxy to an event stream.
    """
    stream: IEventStream

    def is_closed(self):
        """
        Public readonly access to the closed status of the stream. Provided by the stream.
        """
        return self.stream.is_closed()

    def __enter__(self):
        """
        Allows this object to be used as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Closes this stream when used as a context manager.
        """
        self.close()

        if exc_type is not None:
            return False

    def close(self):
        """
        Closes this stream.
        """
        self.stream.close()

    def get_events_to_listen(self):
        """
        Returns the list of events this stream is subscribed to
        """
        return self.stream.get_events_to_listen()

    def get_connection(self) -> 'IConnection':
        """
        Returns the connection this event stream is associated with.
        """
        return self.stream.get_connection()

    async def iterate(self):
        """
        Returns a new async iterator for all recorded events and new events as they are received.
        """
        async for item in self.stream.iterate():
            yield item

    def iterate_already_recorded(self) -> Iterator[_T]:
        """
        Returns an async iterator for all recorded events.
        """
        return self.stream.iterate_already_recorded()



from abc import ABC, abstractmethod
from typing import Coroutine, Any

from aiocdp.core.interfaces.stream import IEventStream, IEventStreamReader


class IConnection(ABC):
    """
    Represents a connection to a CDP target.
    """

    @classmethod
    @abstractmethod
    def init(cls, ws_url: str) -> 'IConnection':
        """
        Initializer method for the IConnection class
        """
        pass

    @abstractmethod
    def is_connected(self) -> bool:
        """
        Returns whether the connection is closed.
        """
        pass

    def close_stream(self, stream: 'IEventStream'):
        """
        Unsubscribes the given event stream from receiving events from the connection.
        """
        pass

    def connect(self) -> Coroutine[None, None, Any]:
        """
        Connects to the websocket and starts the listener task. Returns a future that resolves when the connection is
        established.
        """
        pass

    def disconnect(self) -> Coroutine[None, None, Any]:
        """
        Disconnects from the websocket and cancels the listener task. Returns a future that resolves when the
        connection is closed.
        """
        pass

    def is_stream_closed(self, stream: 'IEventStream') -> bool:
        """
        Checks if the given stream is closed.
        """
        pass

    def open_stream(self, events: list[str]) -> 'IEventStreamReader':
        """
        Opens a new event stream for the given events and returns a reader for it.
        """
        pass

    def send(self, method: str, params: dict) -> Coroutine[None, None, Any]:
        """
        Sends a message to the websocket. Returns a future that resolves when the message is sent.
        """
        pass



from abc import ABC, abstractmethod
from typing import Callable, Any, Generator, Iterator, TypeVar, Generic

from aiocdp.core.interfaces.target import ITarget
from aiocdp.utils import UNDEFINED

_ProcessT = TypeVar('_ProcessT')


class IChrome(ABC, Generic[_ProcessT]):
    """
    Represents an instance of a Chrome Browser
    """

    @classmethod
    @abstractmethod
    def init(
        cls,
        host: str,
        port: int,
    ) -> 'IChrome':
        """
        Initializer method for the Chrome classes
        """
        pass

    @abstractmethod
    def start(
            self,
            command: str = None,
            extra_cli_args: list[str] = None
    ) -> _ProcessT:
        """
        Starts the chrome instance and returns the process.
        """
        pass

    @abstractmethod
    def close_tab(self, target_id: str):
        """
        Closes the tab with the given target id.
        """
        pass

    @abstractmethod
    def get_first_target(
            self,
            condition: Callable[[ITarget], bool] = None,
            default: Any = UNDEFINED
    ) -> ITarget:
        """
        Fetches and returns the first target.
        Raises an exception if no target is found and no default is supplied.
        """
        pass

    @abstractmethod
    def get_host(self) -> str:
        """
        Returns the host of the chrome instance
        """
        pass

    @abstractmethod
    def get_port(self) -> int:
        """
        Returns the port of the chrome instance
        """
        pass

    @abstractmethod
    def get_targets(self) -> list[ITarget]:
        """
        Fetches a list of all the targets.
        """
        pass

    @abstractmethod
    def iterate_targets(self) -> Iterator[ITarget]:
        """
        Returns a generator that fetches and iterates over all the targets.
        """
        pass

    @abstractmethod
    def open_tab(self, url: str = None) -> ITarget:
        """
        Opens a new tab.
        """
        pass



from .core.implementations import (
    Chrome,
    Target,
    Session,
    Connection,
    EventStream,
    EventStreamReader
)
from .core.interfaces import (
    IChrome,
    ITarget,
    ISession,
    IConnection,
    IEventStream,
    IEventStreamReader
)
from .exceptions import (
    AIOCDPException,
    InvalidRPCResponse,
    NoTargetFound,
    NoTargetFoundMatchingCondition
)
from .ioc import (
    get_class,
    set_chrome_class,
    set_target_class,
    set_target_info_class,
    set_session_class,
    set_connection_class,
    set_event_stream_class,
    set_event_stream_reader_class
)
from .setup import (
    setup_default_factories
)



from abc import ABC, abstractmethod
from typing import Any, Coroutine, TYPE_CHECKING

from aiocdp.core.interfaces.stream import IEventStream, IEventStreamReader

if TYPE_CHECKING:
    from aiocdp.core.interfaces.target import ITarget


class ISession(ABC):
    """
    Represents a session with a target.
    """

    @classmethod
    @abstractmethod
    def init(
        cls,
        target: 'ITarget'
    ):
        """
        Initializer method for the ISession class
        """
        pass

    @abstractmethod
    def __aenter__(self):
        """
        Allows this object to be used as an async context manager.
        """
        pass

    @abstractmethod
    def __aexit__(self, exc_type, exc_value, traceback):
        """
        Closes this session when used as an async context manager.
        """
        pass

    @abstractmethod
    def close(self) -> Coroutine[None, None, Any]:
        """
        Closes the session by detaching from the target.
        """
        pass

    @abstractmethod
    def close_stream(self, stream: IEventStream):
        """
        Closes the stream. Calls `Target.close_stream`.
        """
        pass

    @abstractmethod
    def open(self) -> Coroutine[None, None, Any]:
        """
        Opens the session by attaching to the target.
        """
        pass

    @abstractmethod
    def open_stream(self, events: list[str]) -> IEventStreamReader:
        """
        Opens a stream. Calls `self.target.open_stream` method on the target.
        """
        pass

    @abstractmethod
    def send(self, method: str, params: dict = None) -> Coroutine[None, None, Any]:
        """
        Sends a message to the target. Calls `self.target.send`.
        """
        pass

    @abstractmethod
    def send_and_await_response(self, method: str, params: dict = None) -> Coroutine[None, None, Any]:
        """
        Sends a message to the target and awaits a response. Calls `self.target.send_and_await_response`
        """
        pass



from dataclasses import dataclass, field

from aiocdp.ioc import get_class
from aiocdp.core.interfaces.session import ISession
from aiocdp.core.interfaces.connection import IConnection
from aiocdp.core.interfaces.target import ITarget
from aiocdp.core.interfaces.chrome import IChrome
from aiocdp.core.implementations.connection import Connection
from aiocdp.core.implementations.session import Session
from aiocdp.core.implementations.stream import EventStream

from aiocdp.core.interfaces import ITargetInfo


@dataclass
class TargetInfo(ITargetInfo):
    """
    Represents information about a target.
    """

    id: str
    title: str
    description: str
    url: str
    type: str
    web_socket_debugger_url: str
    parent_id: str = None
    favicon_url: str = None

    @classmethod
    def init(
            cls,
            id_: str,
            title: str,
            description: str,
            url: str,
            type_: str,
            web_socket_debugger_url: str,
            parent_id: str = None,
            favicon_url: str = None
    ):
        """
        Initializer method for the ITargetInfo class
        """
        return cls(
            id=id_,
            title=title,
            description=description,
            url=url,
            type=type_,
            web_socket_debugger_url=web_socket_debugger_url,
            parent_id=parent_id,
            favicon_url=favicon_url
        )

    def get_id(self):
        """
        Returns the id of the target.
        """
        return self.id


@dataclass
class Target(ITarget):
    """
    Represents a CDP target.
    """
    _connection: 'IConnection' = field(
        init=False,
        repr=False
    )

    """
    A reference to the parent chrome instance
    """
    chrome: 'IChrome'

    """
    A reference to the target info.
    """
    info: 'ITargetInfo'

    @classmethod
    def init(
        cls,
        chrome: IChrome,
        info: ITargetInfo
    ):
        """
        Initializer method for the ITarget class
        """
        return cls(
            chrome=chrome,
            info=info
        )

    def __post_init__(self):
        """
        Assigns the connection to the target.
        """
        connection_class = get_class(
            IConnection,
            Connection
        )

        self._connection = connection_class.init(self.get_ws_url())

    async def close_session(self, session: 'Session'):
        """
        Closes the session. Calls `TargetSession.close`.
        """
        await session.close()

    def close_stream(self, stream: EventStream):
        """
        Closes the stream. Calls `Connection.close_stream`.
        """
        return self._connection.close_stream(stream)

    async def connect(self):
        """
        Connects to the target. Calls `Connection.connect`.
        """
        return await self._connection.connect()

    async def disconnect(self):
        """
        Disconnects from the target. Calls `Connection.disconnect`.
        """
        return await self._connection.disconnect()

    def get_info(self):
        """
        Returns the associated ITargetInfo implementation
        """
        return self.info

    def get_ws_url(self):
        """
        Returns the associated Web Socket URL
        """
        host = self.chrome.get_host()
        port = self.chrome.get_port()

        target_id = self.info.get_id()

        return f'ws://{host}:{port}/devtools/page/{target_id}'

    def is_connected(self) -> bool:
        """
        Returns whether the target is connected. Calls `Connection.is_connected`.
        """
        return self._connection.is_connected()

    def open_stream(self, events: list[str]):
        """
        Opens a stream. Calls `Connection.open_stream`.
        """
        return self._connection.open_stream(events)

    async def open_session(self):
        """
        Opens a session with the target.
        """
        session_class = get_class(
            ISession,
            Session
        )

        session = session_class.init(self)
        await session.open()

        return session

    async def send(self, method: str, params: dict = None):
        """
        Sends a message to the target. Calls `Connection.send`.
        """
        return await self._connection.send(
            method,
            params or {}
        )

    async def send_and_await_response(self, method: str, params: dict = None):
        """
        Sends a message to the target and awaits a response. Calls `Connection.send_and_await_response`
        """
        return await (await self.send(
            method,
            params
        ))
